--CPML is the magic that makes the transforms work
local cpml = require "cpml"
local modelLoader = require "modelLoader"

local gun = {}
local egg = {}
local cube = {}
local shader, setCamera, makeProper, drawModel

local lightPower = 10
local lightPosition = {3,3,-3}
local lightColor = {1,1,1}

--Function to correct cpml incorrect matrix format.
function makeProper(mat)
	local m = cpml.mat4.new()
	return cpml.mat4.transpose(m, mat)
end

local perspective = makeProper(cpml.mat4.from_perspective(60, love.graphics.getWidth()/love.graphics.getHeight(), 0.1, 1000))

--initial position and angle of Camera
local camera = {
	pos 			= cpml.vec3(0, 0, -15),
	angle			= cpml.vec3(0, 0, 0),
	transform 		= cpml.mat4(),
}


--initial transforms for the model
local gunTransform = cpml.mat4.identity()

local eggTransform = cpml.mat4.identity()
eggTransform:translate(eggTransform, cpml.vec3(12,0,0))

local cubeTransform = cpml.mat4.identity()
cubeTransform:translate(cubeTransform, cpml.vec3(-12,0,0))


--Custom vertex format add z, w to vertex position and the normals for vertices
local format = {
	{
		"VertexPosition",
		"float",
		4
	},
	{
		"VertexTexCoord",
		"float",
		2
	},
	{
		"VertexNormal",
		"float",
		3
	}
}


function love.load()
	love.graphics.setBackgroundColor(.5,.5,.5)

	local gunMesh = love.graphics.newMesh(format, modelLoader("/models/glock.obj"), "triangles")
	gunMesh:setTexture(love.graphics.newImage("/images/carbon.png"))

	gun = {
		transform = gunTransform,
		mesh = gunMesh
	}

	local eggMesh = love.graphics.newMesh(format, modelLoader("/models/egg.obj"), "triangles")
	eggMesh:setTexture(love.graphics.newImage("/images/egg.png"))

	egg = {
		transform = eggTransform,
		mesh = eggMesh
	}

	local cubeMesh = love.graphics.newMesh(format, modelLoader("/models/cube.obj"), "triangles")
	cubeMesh:setTexture(love.graphics.newImage("/images/boxtexture.png"))

	cube = {
		transform = cubeTransform,
		mesh = cubeMesh
	}

	shader = love.graphics.newShader("basicRender.glsl")

	--Sets DepthMode. Prevents items being drawn out of order
	love.graphics.setDepthMode("lequal", true)
	love.graphics.setDefaultFilter("nearest","nearest")

end

function love.draw()

	drawModel(gun)
	drawModel(egg)
	drawModel(cube)

	love.graphics.print("Angle: " .. camera.angle.x .. ", " .. camera.angle.y, 20, 20)
	love.graphics.print("Position: " .. camera.pos.x .. ", " .. camera.pos.y .. ", " .. camera.pos.z, 20, 20*2)
	love.graphics.print("Backspace to reset position and angle", 20, 20*3)
	love.graphics.print("Click TAB to lock mouse", 20, 20*4)
	love.graphics.print("WASDQE to move camera position. RT to change light intensity.", 20, 20*5)
	love.graphics.print("FPS " .. love.timer.getFPS(), 20, 20*6)
end

function love.update(dt)
	local speed = 5 * dt
	local pos = camera.pos
	
	local mul = love.keyboard.isDown("w", "up") and 1 or (love.keyboard.isDown("s", "down") and -1 or 0)
	pos.x = pos.x + math.sin(-camera.angle.x) * mul * speed
	pos.z = pos.z + math.cos(-camera.angle.x) * mul * speed
	
	local mul = love.keyboard.isDown("d", "right") and -1 or (love.keyboard.isDown("a", "left") and 1 or 0)
	pos.x = pos.x + math.cos(camera.angle.x) * mul * speed
	pos.z = pos.z + math.sin(camera.angle.x) * mul * speed

	local mul = love.keyboard.isDown("e") and 1 or (love.keyboard.isDown("q") and -1 or 0)
	pos.y = pos.y + mul * speed

	if love.keyboard.isDown("r") then
		lightPower = lightPower - speed*2
	elseif love.keyboard.isDown("t") then
		lightPower = lightPower + speed*2
	end

end

--Function to merge camera transform and projection matrix(I think anyways)
function getView()
	camera.transform = cpml.mat4() -- Wipe transform
	local t, a, p = camera.transform, camera.angle, camera.pos

	t:rotate(t, a.y, cpml.vec3.unit_x)
	t:rotate(t, a.x, cpml.vec3.unit_y)
	t:rotate(t, a.z, cpml.vec3.unit_z)
	
	t:translate(t, p) -- Rotation and then translating makes the rotation relative to camera, rather than origin
	
	return makeProper(t)
end

function love.keypressed(key)
	if key == "escape" then
		love.event.quit()
	elseif key == "backspace" then
		camera.angle = cpml.vec3(0, 0, 0)
		camera.pos = cpml.vec3(0, 0, -15)
	elseif key == "tab" then
		love.mouse.setRelativeMode(not love.mouse.getRelativeMode())
	end
end

function love.mousemoved(x, y, dx, dy)
	if love.mouse.getRelativeMode() then
		camera.angle.x = camera.angle.x + math.rad(dx * 0.5)
		camera.angle.y = camera.angle.y + math.rad(dy * 0.5)
	end
end

function drawModel(model)
	love.graphics.push("all")
	--centers the world on the screen
	local w, h = love.graphics.getDimensions()
	love.graphics.translate(w/2, h/2)

	--Send MVP to shader
	local modelMatrix = makeProper(model.transform)
	local viewMatrix = getView()
	local MVP = perspective*viewMatrix*modelMatrix
	shader:send("MVP", MVP)
	shader:send("modelMatrix", modelMatrix)
	shader:send("viewMatrix", viewMatrix)
	shader:send("lightPosition", lightPosition)
	shader:send("lightColor", lightColor)
	shader:send("lightPower",lightPower)

	--Sets the shader
	love.graphics.setShader(shader)

	--Draw mesh and clear shader.
	love.graphics.draw(model.mesh)
	love.graphics.setShader()
	love.graphics.pop()
end

